"""Last updated point should not be longer than 24 hours.

<p>This check finds the last time a value was archived. If no value has been recorded in the last
24 hours, the tag might no longer be in use.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
A series that no longer puts out any measurements might be faulty or could indicate a network failure.
Failing to detect this could lead to wrong process operation when attempting to obtain a particular
interval of operation.
</p>
</div>
"""

from datetime import timedelta

import timeseer

_CHECK_NAME = "Recently updated"

META: dict = {
    "checks": [{"name": _CHECK_NAME, "kpi": "Timeliness", "data_type": "bool"}],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
        }
    ],
    "signature": "univariate",
}


def _is_last_point_old(analysis_input) -> bool:
    last_recorded_time = analysis_input.data.index[-1]
    last_evaluation_date = analysis_input.evaluation_time_range.end_date
    cutoff_date = last_evaluation_date - timedelta(days=1)
    return last_recorded_time <= cutoff_date


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if len(analysis_input.data) == 0:
        return timeseer.AnalysisResult(condition_message="No data")

    is_too_old = _is_last_point_old(analysis_input)
    check = timeseer.CheckResult(_CHECK_NAME, float(is_too_old))
    return timeseer.AnalysisResult([check])



"""Timeseer Client provides convenient remote access to Timeseer.

Data, metadata and event frames are exposed as Python objects."""

from kukur import DataType, Dictionary, InterpolationType, Metadata, SeriesSelector

from .internal import (
    AugmentationStrategy,
    ProcessType,
    TimeseerClientException,
    UnknownAugmentationStrategyException,
)
from .internal.client import Client
from .internal.filters import filter_event_frames, filter_series
from .metadata.fields import register_custom_fields


register_custom_fields(Metadata)


__all__ = [
    "AugmentationStrategy",
    "Client",
    "DataType",
    "Dictionary",
    "InterpolationType",
    "Metadata",
    "ProcessType",
    "SeriesSelector",
    "TimeseerClientException",
    "UnknownAugmentationStrategyException",
    "filter_event_frames",
    "filter_series",
]



"""Identification of big jumps in consecutive values.

<p>This check identifies how many times over the time-frame of the analysis a sudden spike occurs.
A sudden spike is defined as an outlier wrt historical differences in consecutive values.</p>
<p><img src='../static/images/reporting/big_jumps.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where a positive or negative jump occurs. Imagine that 100 points are analyzed in a given time-frame
and there are 2 positive and 1 negative jumps. The score for this check in that case would be
97% = 1 - 3 / 100. Which means that for 90% of all points no jump is present.</p>
<div class="ts-check-impact">
<p>
Big sudden jumps in data often correspond to faults in the data captation chain.
Several calculations are sensitive to these type of sudden changes and process decision chains
based on such an outlier can propagate through the system.
Any type of anomaly detection based on normal behavior of the phyiscal sensor can act as a guide for
prioritization of callibration / maintenance.
</p>
</div>
"""

from typing import List

import jsonpickle
import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
    calculate_local_score,
)


_CHECK_NAME = "Jumps"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "event_frames": ["Jump outlier"],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_quantiles(sketch):
    if sketch.max == sketch.min:
        return sketch.max, sketch.max, 0

    q25, q75 = [sketch.get_quantile_value(q) for q in [0.25, 0.75]]
    iqr = q75 - q25
    return q25, q75, iqr


def _get_relevant_jumps(values, jump_up_sketch, jump_down_sketch, strictness):
    ups = values[values > 0]
    down = values[values < 0]

    down_jumps = [False] * len(values)
    up_jumps = [False] * len(values)

    if len(ups) >= 30:
        _, up_q3, up_iqr = _get_quantiles(jump_up_sketch)
        up_jumps = values > (up_q3 + strictness * up_iqr)

    if len(down) >= 30:
        down_q1, _, down_iqr = _get_quantiles(jump_down_sketch)
        down_jumps = values < (down_q1 - strictness * down_iqr)

    if len(ups) < 30 and len(down) < 30:
        return None

    return up_jumps | down_jumps


def _handle_open_intervals(df, intervals):
    if len(intervals) == 0:
        return intervals
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _get_intervals(outliers, df, event_type):
    if outliers is None:
        return pd.DataFrame()
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_jump_check(
    analysis_input,
    median_archival_step,
    jump_up_sketch,
    jump_down_sketch,
    open_event_frames,
    strictness=3,
):  # pylint: disable=too-many-arguments
    df = _clean_dataframe(analysis_input.data)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    value_diff = df["value"].diff()

    active_points = _get_relevant_jumps(
        value_diff, jump_up_sketch, jump_down_sketch, strictness
    )

    intervals = _get_intervals(active_points, df, "Jump outlier")
    intervals = _handle_open_intervals(df, intervals)
    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    percentage = calculate_local_score(
        df, analysis_start, intervals, median_archival_step
    )

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-2]

    return percentage, list(frames), last_analyzed_point


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


def _is_valid_input(analysis_input) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    json_jump_up_sketch = _get_relevant_statistic(analysis_input, "Jump Up Sketch")
    json_jump_down_sketch = _get_relevant_statistic(analysis_input, "Jump Down Sketch")

    if median_archival_step is None:
        return timeseer.AnalysisResult(condition_message="No median archival step")
    if json_jump_up_sketch is None:
        return timeseer.AnalysisResult(condition_message="No jump up sketch")
    if json_jump_down_sketch is None:
        return timeseer.AnalysisResult(condition_message="No jump down sketch")

    jump_up_sketch = jsonpickle.decode(json_jump_up_sketch)
    jump_down_sketch = jsonpickle.decode(json_jump_down_sketch)

    open_event_frames = _get_open_event_frames(analysis_input)

    pct_jumps, event_frames, last_analyzed_point = _run_jump_check(
        analysis_input,
        median_archival_step,
        jump_up_sketch,
        jump_down_sketch,
        open_event_frames,
    )

    if pct_jumps is None:
        return timeseer.AnalysisResult(condition_message="No jump percentage")

    check = timeseer.CheckResult(_CHECK_NAME, float(pct_jumps))
    return timeseer.AnalysisResult(
        check_results=[check],
        event_frames=event_frames,
        last_analyzed_point=last_analyzed_point,
    )



import json
import sqlite3

from typing import Callable

from yoyo import step

from timeseer.blocks import BlockConfiguration, BlockType
from timeseer.blocks.comparison import ComparisonBlock
from timeseer.blocks.data_service_expose import DataServiceExposeBlock
from timeseer.blocks.exports import ExportBlock
from timeseer.blocks.filters import (
    AugmentationStrategy,
    FilterBlock,
    FlowEventFrameFilter,
)
from timeseer.blocks.flight_expose import FlightExposeBlock
from timeseer.blocks.multivariate import MultivariateAnalysisBlock
from timeseer.blocks.trendminer_expose import TrendMinerExposeBlock
from timeseer.blocks.univariate import UnivariateAnalysisBlock
from timeseer.time_ranges import AbsoluteTimeRange


__depends__ = {"044__remove_evaluation_block_fk"}


def migrate_blocks(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute(
        """
        select b.id, b.flow_id, b.sort_order, c.id, c.name, c.type
        from Blocks b, BlockConfigurations c
        where b.configuration_id = c.id
    """
    )

    configuration_fns: dict[
        BlockType, Callable[[sqlite3.Cursor, int], BlockConfiguration]
    ] = {
        BlockType.COMPARISON: get_comparison_block,
        BlockType.DATA_SERVICE_EXPOSE: get_data_service_expose_block,
        BlockType.EXPORT: get_export_block,
        BlockType.FILTER: get_filter_block,
        BlockType.FLIGHT_EXPOSE: get_flight_expose_block,
        BlockType.MULTIVARIATE_ANALYSIS: get_multivariate_block,
        BlockType.TRENDMINER_EXPOSE: get_trendminer_expose_block,
        BlockType.UNIVARIATE_ANALYSIS: get_univariate_block,
    }

    for (
        block_id,
        flow_id,
        sort_order,
        configuration_id,
        block_name,
        block_type,
    ) in cursor.fetchall():
        configuration = configuration_fns[BlockType[block_type]](
            cursor, configuration_id
        )
        effective_name = block_name
        counter = 0
        while True:
            try:
                cursor.execute(
                    """
                    insert into Blocks_new (id, flow_id, type, name, configuration, sort_order)
                    values (?, ?, ?, ?, ?, ?)
                """,
                    [
                        block_id,
                        flow_id,
                        block_type,
                        effective_name,
                        json.dumps(configuration.to_data()),
                        sort_order,
                    ],
                )
                break
            except sqlite3.IntegrityError as err:
                counter = counter + 1
                if counter > 5:
                    raise err
                effective_name = f"{block_name} ({counter})"


steps = [
    step(
        """
            create table Blocks_new (
                id integer primary key autoincrement,
                flow_id integer not null,
                type text not null,
                name text not null,
                configuration text not null,
                sort_order integer not null,

                unique (flow_id, name),
                foreign key (flow_id) references Flows(id) on delete cascade
            );
        """,
    ),
    step(migrate_blocks),
    step("drop table Blocks"),
    step("alter table Blocks_new rename to Blocks"),
    step(
        """
            create table FlowDataService_new (
                id integer primary key autoincrement,
                name text not null unique,
                block_evaluation_id integer not null,
                flow_id integer not null,
                series_set_name text,
                block_name text not null,

                foreign key (block_evaluation_id) references BlockEvaluations(id) on delete cascade,
                foreign key (flow_id) references Flows(id) on delete cascade
            );
    """
    ),
    step(
        """
        insert into FlowDataService_new (id, name, block_evaluation_id, flow_id, series_set_name, block_name)
        select d.id, d.name, d.block_evaluation_id, d.flow_id, d.series_set_name, c.name
    from FlowDataService d, BlockConfigurations c
        where d.block_configuration_id = c.id
    """
    ),
    step("drop table FlowDataService"),
    step("alter table FlowDataService_new rename to FlowDataService"),
    step(
        """
            create table FlowSource_new (
                id integer primary key autoincrement,
                flow_id integer not null,
                series_set_name text not null,
                block_name text not null,
                block_evaluation_id integer not null,
                name text not null unique,

                foreign key (block_evaluation_id) references BlockEvaluations(id) on delete cascade,
                foreign key (flow_id) references Flows(id) on delete cascade
            );
    """
    ),
    step(
        """
        insert into FlowSource_new (id, name, block_evaluation_id, flow_id, series_set_name, block_name)
        select f.id, f.name, f.block_evaluation_id, f.flow_id, f.series_set_name, c.name
        from FlowSource f, BlockConfigurations c
        where f.block_configuration_id = c.id
    """
    ),
    step("drop table FlowSource"),
    step("alter table FlowSource_new rename to FlowSource"),
    step("drop table BlockConfigurations"),
    step("drop table ComparisonBlockInput"),
    step("drop table DataServiceExposeBlockInput"),
    step("drop table ExportBlockInput"),
    step("drop table FilterBlockInput"),
    step("drop table FlightExposeBlockInput"),
    step("drop table MultivariateBlockInput"),
    step("drop table TrendMinerExposeBlockInput"),
    step("drop table UnivariateBlockInput"),
]


def get_comparison_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> ComparisonBlock:
    """get the comparison block."""
    cursor.execute(
        """
        select reference_start_date, reference_end_date from ComparisonBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    reference_start_date, reference_end_date = cursor.fetchone()
    return ComparisonBlock(AbsoluteTimeRange(reference_start_date, reference_end_date))


def get_data_service_expose_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> DataServiceExposeBlock:
    """Get the data service expose block."""
    cursor.execute(
        """
        select pattern from DataServiceExposeBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    (pattern,) = cursor.fetchone()
    return DataServiceExposeBlock(pattern)


def get_export_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> ExportBlock:
    """get the report block."""
    cursor.execute(
        """
        select export from ExportBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    exports = [export for export, in cursor]
    return ExportBlock(exports)


def get_filter_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> FilterBlock:
    """get the filter block."""
    cursor.execute(
        """
        select frame_type, frame_type_category, augmentation_strategy, series_id, filter_selection_only
        from FilterBlockInput
        where block_id = ?
    """,
        [block_configuration_id],
    )
    filters = []
    augmentation_strategy = None
    for (
        frame_type,
        frame_type_category,
        strategy,
        series_id,
        filter_selection_only,
    ) in cursor:
        augmentation_strategy = AugmentationStrategy[strategy]
        filters.append(
            FlowEventFrameFilter.from_string(
                frame_type,
                frame_type_category,
                series_id,
                bool(filter_selection_only),
            )
        )
    return FilterBlock(filters, augmentation_strategy)


def get_flight_expose_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> FlightExposeBlock:
    """Return a flight expose block."""
    cursor.execute(
        """
        select pattern from FlightExposeBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    (pattern,) = cursor.fetchone()
    return FlightExposeBlock(pattern)


def get_multivariate_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> MultivariateAnalysisBlock:
    """get the multivariate analysis block."""
    cursor.execute(
        """
        select module_type from MultivariateBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    module_types = [module_type for module_type, in cursor]
    return MultivariateAnalysisBlock(module_types, [])


def get_trendminer_expose_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> TrendMinerExposeBlock:
    """get the TrendMiner expose block."""
    cursor.execute(
        """
        select name, prefix from TrendMinerExposeBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    name, prefix = cursor.fetchone()
    return TrendMinerExposeBlock(name, prefix)


def get_univariate_block(
    cursor: sqlite3.Cursor, block_configuration_id: int
) -> UnivariateAnalysisBlock:
    """get the univariate block."""
    cursor.execute(
        """
        select module_type from UnivariateBlockInput
        where block_configuration_id = ?
    """,
        [block_configuration_id],
    )
    module_types = [module_type for module_type, in cursor]
    return UnivariateAnalysisBlock(module_types, [])



"""Anomaly based on level shift.

<p>Anomaly detection based on a significant difference in the means of
two consecutive sliding windows.</p>
<p><img src='../static/images/reporting/level_shift.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where a level shift anomaly occurs. Imagine that 100 points are analyzed in a given time-frame
and there are 10 level shift anomalies. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no level shift anomaly occurs.</p>
<div class="ts-check-impact">
<p>
Level shifts could be indications of physical altercations to a sensor (e.g. camera being bumped) or
other types of issues within the sensor.
Several calculations are sensitive to these type of sudden changes and process decision chains
based on such an outlier can propagate through the system.
Any type of anomaly detection based on normal behavior of the phyiscal sensor can act as a guide for
prioritization of callibration / maintenance.
</p>
</div>
"""

import logging

from typing import List

import numpy as np
import pandas as pd

from adtk.detector import LevelShiftAD
from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType, InterpolationType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
    calculate_local_score,
)


_CHECK_NAME = "Level Shift Anomaly"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "event_frames": ["Level Shift Anomaly"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
            "interpolation_type": [None, InterpolationType.LINEAR],
        }
    ],
    "signature": "univariate",
}

logger = logging.getLogger(__name__)


def _get_intervals(active_points, df, event_type):
    interval_grp = (active_points != active_points.shift().bfill()).cumsum()
    active_points[active_points.isna()] = 0
    active_points = np.array(active_points, dtype=bool)
    intervals = (
        df.assign(interval_grp=interval_grp)[active_points]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _get_last_analyzed_point(df, window_size):
    if len(df) < window_size:
        return df.index[0]
    last_index = -2 * window_size
    return df.index[last_index]


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_level_shift_ad(analysis_input, median_archival_step, open_event_frames):
    df = _clean_dataframe(analysis_input.data)

    window_size = 10
    if len(df["value"]) <= window_size:
        return None, pd.DataFrame(), df.index[0]

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    level_ad = LevelShiftAD(window=10)
    try:
        active_points = level_ad.fit_detect(df["value"])
    except (ValueError, RuntimeError) as err:
        logger.error("Error in level shift anomaly: %s", err)
        return None, pd.DataFrame(), _get_last_analyzed_point(df, window_size)

    intervals = _get_intervals(active_points, df, _CHECK_NAME)
    intervals = _handle_open_intervals(df, intervals)
    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    percentage = calculate_local_score(
        df, analysis_start, intervals, median_archival_step
    )

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = _get_last_analyzed_point(df, window_size)

    return percentage, list(frames), last_analyzed_point


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _get_relevant_statistic(analysis_input, requested_statistic):
    stat = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == requested_statistic
    ]
    if len(stat) == 0:
        return None
    return stat


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    if median_archival_step is None:
        return timeseer.AnalysisResult(condition_message="No median archival step")

    open_event_frames = _get_open_event_frames(analysis_input)

    pct_anomaly, frames, last_analyzed_point = _run_level_shift_ad(
        analysis_input, median_archival_step[0], open_event_frames
    )

    if pct_anomaly is None:
        return timeseer.AnalysisResult(condition_message="No anomaly percentage")

    check = timeseer.CheckResult(_CHECK_NAME, float(pct_anomaly))
    return timeseer.AnalysisResult(
        [check], frames, last_analyzed_point=last_analyzed_point
    )



import json
import sqlite3

from yoyo import step

from timeseer.blocks import BlockType

__depends__ = {"045__remove_block_configurations"}


def update_blocks(db: sqlite3.Connection):
    cursor = db.cursor()
    _update_univariate_blocks(cursor)
    _update_univariate_block_evaluations(cursor)
    _update_multivariate_blocks(cursor)
    _update_multivariate_block_evaluations(cursor)


def _update_univariate_blocks(cursor: sqlite3.Cursor):
    cursor.execute(
        """
        select id, configuration
        from Blocks
        where type = ?
    """,
        [BlockType.UNIVARIATE_ANALYSIS.name],
    )
    for db_id, configuration in cursor.fetchall():
        new_configuration = _build_univariate_configuration(configuration)
        cursor.execute(
            """
            update Blocks
            set configuration = ?
            where id = ?
        """,
            [json.dumps(new_configuration), db_id],
        )


def _update_univariate_block_evaluations(cursor: sqlite3.Cursor):
    cursor.execute(
        """
        select id, configuration
        from BlockEvaluations
        where type = ?
    """,
        [BlockType.UNIVARIATE_ANALYSIS.name],
    )
    for db_id, configuration in cursor.fetchall():
        new_configuration = _build_univariate_configuration(configuration)
        cursor.execute(
            """
            update BlockEvaluations
            set configuration = ?
            where id = ?
        """,
            [json.dumps(new_configuration), db_id],
        )


def _build_univariate_configuration(configuration):
    configuration = json.loads(configuration)
    new_configuration = []
    for module_type in configuration["moduleTypes"]:
        if module_type == "metadata_bugs":
            new_configuration.append("metadata_statistics")
            if "metadata_checks" not in new_configuration:
                new_configuration.append("metadata_checks")
        elif module_type == "metadata_smells":
            if "metadata_checks" not in new_configuration:
                new_configuration.append("metadata_checks")
        elif module_type in ["univariate_bugs", "univariate_smells"]:
            if "univariate_checks" not in new_configuration:
                new_configuration.append("univariate_checks")
        else:
            new_configuration.append(module_type)
    configuration["moduleTypes"] = new_configuration
    return configuration


def _update_multivariate_blocks(cursor: sqlite3.Cursor):
    cursor.execute(
        """
        select id, configuration
        from Blocks
        where type = ?
    """,
        [BlockType.MULTIVARIATE_ANALYSIS.name],
    )
    for db_id, configuration in cursor.fetchall():
        new_configuration = _build_multivariate_configuration(configuration)
        cursor.execute(
            """
            update Blocks
            set configuration = ?
            where id = ?
        """,
            [json.dumps(new_configuration), db_id],
        )


def _update_multivariate_block_evaluations(cursor: sqlite3.Cursor):
    cursor.execute(
        """
        select id, configuration
        from BlockEvaluations
        where type = ?
    """,
        [BlockType.MULTIVARIATE_ANALYSIS.name],
    )
    for db_id, configuration in cursor.fetchall():
        new_configuration = _build_multivariate_configuration(configuration)
        cursor.execute(
            """
            update BlockEvaluations
            set configuration = ?
            where id = ?
        """,
            [json.dumps(new_configuration), db_id],
        )


def _build_multivariate_configuration(configuration):
    configuration = json.loads(configuration)
    new_configuration = []
    for module_type in configuration["moduleTypes"]:
        if module_type in ["multivariate_bugs", "multivariate_smells"]:
            if "multivariate_checks" not in new_configuration:
                new_configuration.append("multivariate_checks")
        else:
            new_configuration.append(module_type)
    configuration["moduleTypes"] = new_configuration
    return configuration


steps = [
    step(update_blocks),
]



import sqlite3

from yoyo import step

__depends__ = {"045__remove_block_configurations"}


def update_block_evaluations(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.executescript(
        """
        alter table BlockEvaluations add column type text;
        alter table BlockEvaluations add column name text;
        alter table BlockEvaluations add column configuration text;
        alter table BlockEvaluations add column sort_order integer;

        create unique index block_evaluations_fe_name on BlockEvaluations(flow_evaluation_id, name);
    """
    )

    cursor.execute("select id, block_id from BlockEvaluations")
    for block_evaluation_id, block_id in cursor.fetchall():
        cursor.execute(
            """
            select type, name, configuration, sort_order
            from Blocks
            where id = ?
        """,
            [block_id],
        )
        block_type, name, configuration, sort_order = cursor.fetchone()
        cursor.execute(
            """
            update BlockEvaluations
            set
                type = ?,
                name = ?,
                configuration = ?,
                sort_order = ?
            where id = ?
        """,
            [block_type, name, configuration, sort_order, block_evaluation_id],
        )


steps = [
    step(update_block_evaluations),
]



"""Dictionary labels should not be repeated with different values.

<p>Dictionaries, or 'digital sets' label discrete numerical values with a
textual representation.
For example, a valve could be 'OPEN' or 'CLOSED'.
'OPEN' could be stored as 1 and 'CLOSED' as 2.<p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Multiple numerical values should not be labeled the same.
This confuses analytics tooling that operates on numerical values, but allows
selection based on labels.</p>
</div>
"""

from collections import Counter

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Labels are not duplicated"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Data types",
            "data_type": "bool",
        }
    ],
    "conditions": [{"min_series": 1, "data_type": [DataType.DICTIONARY]}],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    dictionary = analysis_input.metadata.get_field(fields.Dictionary)
    if dictionary is None:
        return timeseer.AnalysisResult(condition_message="No dictionary")

    counter = Counter(v for v in dictionary.mapping.values())
    has_duplicates = any(count > 1 for count in counter.values())

    return timeseer.AnalysisResult(
        check_results=[
            timeseer.CheckResult(_CHECK_NAME, float(has_duplicates)),
        ],
    )



"""Identification of sensor difference outliers between 2 series.

<p>This check identifies periods where for redundant sensors the typical
difference is significantly altered compared to normal.</p>
<p><img src='../static/images/reporting/sensor_outlier.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where sensor drift is identified. Imagine that 100 points are analyzed in a given time-frame
and that drift is detected for 10 (consecutive) points. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no drift occurs.</p>
<div class="ts-check-impact">
<p>Changes in a physical relation between a set of series could indicate process or instrumentation issues.</p>
</div>
"""

from datetime import timedelta

import pandas as pd
import numpy as np

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Redundant sensor outlier"
_EVENT_FRAME_NAME = "Sensor outlier"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Accuracy",
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 2,
            "max_series": 2,
        },
        {
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        },
    ],
    "signature": "multivariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _is_frame_long_enough(frame):
    return (frame.end_date - frame.start_date) >= timedelta(hours=12)


def _filter_stale_event_frames(all_frames):
    filter_iterator = filter(_is_frame_long_enough, all_frames)
    return filter_iterator


def _get_anomalies(lower, upper, arr):
    a_lower = arr < lower
    a_upper = arr > upper
    return a_lower | a_upper


def _find_anomalous_times(diff_matrix):
    q25, q75 = np.nanquantile(np.array(diff_matrix, dtype=float), [0.25, 0.75], axis=0)
    iqr = q75 - q25
    upper = q75 + 1.5 * iqr
    lower = q25 - 1.5 * iqr
    return [_get_anomalies(lower, upper, x) for x in np.array(diff_matrix, dtype=float)]


def _make_diff_matrix(df):
    diff_matrix = np.zeros((df.shape[0], df.shape[1], df.shape[1]))
    for i in range(df.shape[1] - 1):
        for j in range(i + 1, df.shape[1]):
            diff_matrix[:, i, j] = df.iloc[:, i] - df.iloc[:, j]
    return diff_matrix


def _anomalous_times_per_series(anomalous_times_matrix, series):
    results = []
    for i in range(len(series) - 1):
        for j in range(i + 1, len(series)):
            results.append(
                {
                    "series_x": series[i],
                    "series_y": series[j],
                    "drifts": [time[i, j] for time in anomalous_times_matrix],
                }
            )
    return results


def _get_intervals(anomalies, df):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME
    return intervals


def _get_results_per_series(
    series_anomaly, df, median_archival_step
) -> timeseer.BivariateCheckResult:
    anomalies = series_anomaly["drifts"]
    intervals = _get_intervals(anomalies, df)

    check_active_time = _calculate_total_check_active_time(
        intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(intervals)

    total_time = df.index[-1] - df.index[0]
    percentage = check_active_time / total_time

    return timeseer.BivariateCheckResult(
        _CHECK_NAME,
        series_anomaly["series_x"],
        series_anomaly["series_y"],
        percentage,
        _filter_stale_event_frames(frames),
    )


def _clean_input(inputs):
    concatenated_df = (
        pd.concat(
            [
                series.data["value"][
                    ~series.data["value"].index.duplicated(keep="first")
                ]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("linear")
        .dropna()
    )
    concatenated_df = concatenated_df[(concatenated_df != 0).all(1)]
    return concatenated_df


def _get_mv_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _run_detect_redundant_sensor_outlier(inputs):
    concatenated_df = _clean_input(inputs)
    median_archival_step = _get_mv_median_archival_step(concatenated_df)

    diff_matrix = _make_diff_matrix(concatenated_df)
    anomalous_times_matrix = _find_anomalous_times(diff_matrix)

    series = [series.metadata.series for series in inputs]
    anomalous_times_per_series = _anomalous_times_per_series(
        anomalous_times_matrix, series
    )

    return [
        _get_results_per_series(series_anomaly, concatenated_df, median_archival_step)
        for series_anomaly in anomalous_times_per_series
    ]


def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = analysis_input.inputs

    results = _run_detect_redundant_sensor_outlier(inputs)

    return timeseer.AnalysisResult(bivariate_check_results=results)



"""Analysis on the variation in the archival time distribution.

<p>Overcompression is a consequence of settings in a compression algorithm that
cause the archived data to no longer represent the real system accurately. Within traditional
historians this is typically caused by setting the values for exception and/or compression too high.
Low variation in inter-archival time could indicate that the max-time for archival
is often reached.</p>
<p><img src='../static/images/reporting/flat_archival.svg'></p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Badly compressed data, specifically overcompression,
can lead to critical events such as upsets, safety issues and downtime.
</p>
</div>
"""

import numpy as np
import pandas as pd

import jsonpickle

import timeseer

_CHECK_NAME = "Compression - flat archival rate"

META = {
    "checks": [{"name": _CHECK_NAME, "kpi": "Accuracy", "data_type": "bool"}],
    "conditions": [{"min_series": 1, "min_weeks": 1, "min_data_points": 300}],
    "signature": "univariate",
}


def _is_sampling_rate_too_regular(sketch):
    q1, q9 = [sketch.get_quantile_value(q) for q in [0.1, 0.9]]
    if np.isnan(q1) or np.isnan(q9):
        return 0
    if q1 == q9:
        return 1
    return 0


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_flat_check(archival_sketch):
    return _is_sampling_rate_too_regular(archival_sketch)


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    json_archival_sketch = _get_relevant_statistic(analysis_input, "Archival Sketch")
    if json_archival_sketch is None:
        return timeseer.AnalysisResult(condition_message="No archival sketch")
    archival_sketch = jsonpickle.decode(json_archival_sketch)

    score = _run_flat_check(archival_sketch)

    check = timeseer.CheckResult(_CHECK_NAME, float(score))
    return timeseer.AnalysisResult([check])



"""Frequency of sampling / archival rate should be similar for most days.

<p>In general the variation in most processes remains reasonably stable.
Under stable conditions the inter-archival time should also remain similar.
Sensor issues could cause archival rate to become more fluctuant.</p>
<p><img src='../static/images/reporting/frequency.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the sum of time of all
the event-frames in which the archival frequency is considered an outlier compared to the overall mean
time between archivals.
E.g. assume a total period being
analyzed of 1 year and 2 event-frames of 1 month and 2 months respectively. The score of this check
will then be 75% = 1 - 3 / 12. Which means that in 75% of time no variance drift is detected.</p>
<div class="ts-check-impact">
<p>
Frequency changes in archival rate might be caused by changing compression settings which could
lead to an infidel representation of the system. Another possibility is temporary disconnects or outages
of the sensor / network.
This could also be caused by network attacks where data is being manipulated before going to the historian.
</p>
</div>
"""

from datetime import timedelta
from typing import List

import pandas as pd
import numpy as np

import timeseer

from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
    calculate_local_score,
)

_CHECK_NAME = "Consistent daily archival frequency"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Consistency",
            "event_frames": ["Daily frequency outlier"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 4,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


def _get_median_diff_times(df):
    # If only one value in a day, we assume 1 sample per day
    if len(df) <= 1:
        return 24 * 60 * 60

    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times = diff_times.dropna()
    diff_times = diff_times.apply(timedelta.total_seconds)
    return diff_times.median()


def _get_daily_diff_times(df):
    resampled_daily_diff_medians = df.resample("1D").apply(_get_median_diff_times)
    return resampled_daily_diff_medians


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["end_date"] = outlier_intervals["end_date"] + timedelta(days=1)
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] >= df.index[-1]
    return intervals


def _get_quantiles_and_iqr(diff_times):
    q25, q75 = np.nanquantile(diff_times, [0.25, 0.75])
    iqr = q75 - q25
    return q25, q75, iqr


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_daily_frequency_outlier_check(analysis_input, open_event_frames):
    df = _clean_dataframe(analysis_input.data)
    df_daily = df.resample("1D").mean()

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    diff_times_resampled = _get_daily_diff_times(df)["value"].values

    q25, q75, iqr = _get_quantiles_and_iqr(diff_times_resampled)
    outliers = (diff_times_resampled > (q75 + 3 * iqr)) | (
        diff_times_resampled < (q25 - 3 * iqr)
    )

    intervals = _get_intervals(
        outliers, df.resample("D").asfreq(), "Daily frequency outlier"
    )
    intervals = _handle_open_intervals(df_daily, intervals)
    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    percentage = calculate_local_score(df_daily, analysis_start, intervals)

    all_frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df_daily.index[-1]

    return percentage, list(all_frames), last_analyzed_point


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> tuple[str, bool]:
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data frame", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    open_event_frames = _get_open_event_frames(analysis_input)

    (
        pct_inconcistent_freq,
        frames,
        last_analyzed_point,
    ) = _run_daily_frequency_outlier_check(analysis_input, open_event_frames)

    if pct_inconcistent_freq is None:
        return timeseer.AnalysisResult(
            condition_message="No inconsistency frequency percentage"
        )

    check = timeseer.CheckResult(_CHECK_NAME, float(pct_inconcistent_freq))
    return timeseer.AnalysisResult(
        [check], frames, last_analyzed_point=last_analyzed_point
    )



"""The series should have upper operating limits.

<p>Each sensor or lab measuring equipment has an inherent measurement range. Outside of this range
the readings from the sensor can't be relied on. This information is typically present in the
sensor specification sheet.
Traditional historians also allow setting limits (often Low-Low, Low, High and High-High limits).
These limits are often based on process safety concerns or indeed taken from the normal operating range.
In all cases having access to these limits is essential for interpreting readings from the equipment.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check belongs to the score group Unit.
If this check is True, it means that the score for group Unit will be 0%.</p>
<div class="ts-check-impact">
<p>Missing or badly configured limits can cause missing warning signs of abnormal situations both
in the process as well as with the instrumentation.
</p>
</div>
"""

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Upper limit is present"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Limits",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {"min_series": 1, "data_type": [DataType.FLOAT32, DataType.FLOAT64]}
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    limit_high = analysis_input.metadata.get_field(fields.LimitHighPhysical)

    if limit_high is None:
        score = 1.0
    else:
        score = 0.0
    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, score)]
    )



"""Upper limit set in the metadata should not be exceeded.

<p>This check identifies whether the upper limit, if defined, is crossed
within the analyzed time frame.</p>
<p><img src='../static/images/reporting/limits.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where the corresponding value is higher than the given limit.
Imagine that 100 points are analyzed in a given time-frame
and there are 10 points whose value is above the given limit.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that 90% of all points lie below the limit.</p>
<div class="ts-check-impact">
<p>
When the sensor spec limits are exceeded this is an indication of sensor failure. This might mean the
sensor needs to be recalibrated.
</p>
</div>
"""

from datetime import timedelta
from typing import List

import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe
from timeseer.metadata import fields

_CHECK_NAME = "Out-of-bounds (upper, physical)"
_EVENT_FRAME_NAME = "Out of bounds (upper, physical)"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame, metadata: timeseer.Metadata):
    return df["value"] > metadata.get_field(fields.LimitHighPhysical)


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_limit_check(
    metadata: timeseer.Metadata, df: pd.DataFrame, median_archival_step
):
    df = _clean_dataframe(df)

    active_points = _get_active_points(df, metadata)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    check_active_time = _calculate_total_check_active_time(
        intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(intervals)

    total_time = df.index[-1] - df.index[0]
    percentage = check_active_time / total_time

    return percentage, list(frames)


def _is_valid_input(
    analysis_input: timeseer.AnalysisInput, median_archival_step: List[float]
) -> tuple[str, bool]:
    if analysis_input.metadata.get_field(fields.LimitHighPhysical) is None:
        return "No physical upper limit", False
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival step", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: timeseer.AnalysisInput) -> timeseer.AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    message, is_ok = _is_valid_input(analysis_input, median_archival_step)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    pct_active, frames = _run_limit_check(
        analysis_input.metadata, analysis_input.data, median_archival_step[0]
    )
    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult([check], frames)



"""Timeseer score module that generates scores for multivariate checks."""

from typing import Any, Dict

from timeseer import ScoreInput, ScoreOutput, ScoreResult

META: Dict[str, Any] = {
    "scores": [],  # accept all checks
}


# pylint: disable=missing-function-docstring
def score(
    score_input: ScoreInput,
) -> ScoreOutput:
    return ScoreOutput(
        [
            ScoreResult(result.name, result.result)
            for result in score_input.check_results
        ]
    )



import os
from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [line.strip() for line in open("requirements.txt").readlines()]
requirements.extend([line.strip() for line in open("requirements-unix.txt").readlines()])

setup(
    name="timeseer",
    version=os.environ.get('TIMESEER_VERSION', '0.0.0'),
    author="Timeseer.AI",
    author_email="pypi@timeseer.ai",
    description="Timeseer.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    python_requires='>=3.9',
    include_package_data=True,
)



"""Counters should increase in value (until reset).

<p>
This check identifies when a counter goes down for a period.
</p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
time where there is a downward trend.
</p>
<div class="ts-check-impact">
<p>
Could impact downstream dashboards, reports or automated analytics.
</p>
</div>
"""
from typing import List

import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    calculate_local_score,
    process_open_intervals,
)

_CHECK_NAME = "Counter increase"
_EVENT_FRAME_NAME = "Counter decrease"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [
                DataType.FLOAT64,
                DataType.FLOAT32,
            ],
        }
    ],
    "signature": "univariate",
}


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame):
    return df["value"].diff() < 0


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _run_increase_check(analysis_input, median_archival_step, open_event_frames):
    df = analysis_input.data
    df = _clean_dataframe(df)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    active_points = _get_active_points(df)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = _handle_open_intervals(df, intervals)

    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    percentage = calculate_local_score(
        df, analysis_start, intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-1]

    return percentage, list(frames), last_analyzed_point


def _is_valid_input(
    analysis_input: timeseer.AnalysisInput, median_archival_step: List[float]
) -> tuple[str, bool]:
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival step", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No clean data", False
    return "OK", True


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


# pylint: disable=missing-function-docstring
def run(analysis_input: timeseer.AnalysisInput) -> timeseer.AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    message, is_ok = _is_valid_input(analysis_input, median_archival_step)
    if not is_ok:
        return timeseer.AnalysisResult(condition_message=message)

    open_event_frames = _get_open_event_frames(analysis_input)

    pct_active, frames, last_analyzed_point = _run_increase_check(
        analysis_input, median_archival_step[0], open_event_frames
    )
    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult(
        [check], frames, last_analyzed_point=last_analyzed_point
    )



"""Deadband should not be too big compared to range of tag.

<p>This check evaluates the ratio of the deadband as given in the metadata
to the total range as given in the metadata. Typical settings for this measurement are
in the range of [0.5-2]%</p>
<div class="ts-check-impact">
<p>A too high ratio, indicating a high deadband with regards to the typical values of the series
is an indication of compression issues. Badly compressed data, specifically overcompression,
can lead to critical events such as upsets, safety issues and downtime.</p>
</div>
"""

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Deadband to range ratio"

META = {
    "checks": [
        {
            "name": "Deadband to range ratio",
            "kpi": "Metadata",
            "group": "Compression",
        }
    ],
    "conditions": [
        {"min_series": 1, "data_type": [DataType.FLOAT32, DataType.FLOAT64]}
    ],
    "signature": "univariate",
}


def _get_deadband_to_range_ration(
    deadband: float, limit_low: float, limit_high: float
) -> float:
    return min([deadband / (limit_high - limit_low), 1])


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    metadata = analysis_input.metadata
    accuracy = metadata.get_field(fields.Accuracy)
    limit_low = metadata.get_field(fields.LimitLowFunctional)
    limit_high = metadata.get_field(fields.LimitHighFunctional)

    if accuracy is None:
        return timeseer.AnalysisResult(condition_message="No accuracy")
    if accuracy < 0:
        return timeseer.AnalysisResult(condition_message="Negative accuracy")
    if limit_low is None:
        return timeseer.AnalysisResult(condition_message="No functional lower limit")
    if limit_high is None:
        return timeseer.AnalysisResult(condition_message="No functional upper limit")
    if limit_high <= limit_low:
        return timeseer.AnalysisResult(
            condition_message="Functional upper limit is lower than the functional lower limit"
        )

    ratio = _get_deadband_to_range_ration(accuracy, limit_low, limit_high)
    check = timeseer.CheckResult(_CHECK_NAME, float(ratio))
    return timeseer.AnalysisResult([check])



""" Calculate the distributed staleness sketch of the time series. """

import jsonpickle
from ddsketch.ddsketch import DDSketch

import numpy as np

import timeseer


META: dict = {
    "statistics": [
        {"name": "Stale Sketch"},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


def _get_repeated_lengths(df):
    active_points = df["value"].eq(df["value"].shift().bfill()).astype(bool)
    interval_grp = (active_points != active_points.shift().bfill()).cumsum()
    active_points[active_points.isna()] = 0
    active_points = np.array(active_points, dtype=bool)
    intervals = (
        df.assign(interval_grp=interval_grp)[active_points]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    return intervals


def _clean_dataframe(df):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_get_sketch(analysis_input, sketch):
    df = _clean_dataframe(analysis_input.data)
    if len(df) == 0:
        return sketch

    if sketch is None:
        sketch = DDSketch(0.001)

    repeated_lengths = _get_repeated_lengths(df)
    for _, row in repeated_lengths.iterrows():
        sketch.add((row["end_date"] - row["start_date"]).total_seconds())

    return sketch


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return jsonpickle.decode(statistics[0])


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    previous_sketch = _get_relevant_statistic(analysis_input, "Stale Sketch")

    sketch = _run_get_sketch(analysis_input, previous_sketch)
    if sketch is None:
        return timeseer.AnalysisResult(condition_message="No stale sketch")

    sketch_statistic = timeseer.Statistic(
        META["statistics"][0]["name"], "sketch", jsonpickle.encode(sketch)
    )
    return timeseer.AnalysisResult(statistics=[sketch_statistic])
